package ru.dbotthepony.kommons.io

import ru.dbotthepony.kommons.util.DelegateGetter
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2f
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kommons.vector.Vector3d
import ru.dbotthepony.kommons.vector.Vector3f
import ru.dbotthepony.kommons.vector.Vector3i
import ru.dbotthepony.kommons.vector.Vector4d
import ru.dbotthepony.kommons.vector.Vector4f
import ru.dbotthepony.kommons.vector.Vector4i
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream

private inline fun <T, I> InputStream.read2(factory: (I, I) -> T, reader: (InputStream) -> I): T {
	return factory(reader(this), reader(this))
}

private inline fun <T, I> InputStream.read3(factory: (I, I, I) -> T, reader: (InputStream) -> I): T {
	return factory(reader(this), reader(this), reader(this))
}

private inline fun <T, I> InputStream.read4(factory: (I, I, I, I) -> T, reader: (InputStream) -> I): T {
	return factory(reader(this), reader(this), reader(this), reader(this))
}

fun InputStream.readVector2i() = read2(::Vector2i) { it.readInt() }
fun InputStream.readVector2d() = read2(::Vector2d) { it.readDouble() }
fun InputStream.readVector2f() = read2(::Vector2f) { it.readFloat() }

fun InputStream.readVector3i() = read3(::Vector3i) { it.readInt() }
fun InputStream.readVector3d() = read3(::Vector3d) { it.readDouble() }
fun InputStream.readVector3f() = read3(::Vector3f) { it.readFloat() }

fun InputStream.readVector4i() = read4(::Vector4i) { it.readInt() }
fun InputStream.readVector4d() = read4(::Vector4d) { it.readDouble() }
fun InputStream.readVector4f() = read4(::Vector4f) { it.readFloat() }

val Vector2iCodec = StreamCodec.Impl(DataInputStream::readVector2i, DataOutputStream::writeStruct2i)
val Vector2dCodec = StreamCodec.Impl(DataInputStream::readVector2d, DataOutputStream::writeStruct2d)
val Vector2fCodec = StreamCodec.Impl(DataInputStream::readVector2f, DataOutputStream::writeStruct2f)

val Vector3iCodec = StreamCodec.Impl(DataInputStream::readVector3i, DataOutputStream::writeStruct3i)
val Vector3dCodec = StreamCodec.Impl(DataInputStream::readVector3d, DataOutputStream::writeStruct3d)
val Vector3fCodec = StreamCodec.Impl(DataInputStream::readVector3f, DataOutputStream::writeStruct3f)

val Vector4iCodec = StreamCodec.Impl(DataInputStream::readVector4i, DataOutputStream::writeStruct4i)
val Vector4dCodec = StreamCodec.Impl(DataInputStream::readVector4d, DataOutputStream::writeStruct4d)
val Vector4fCodec = StreamCodec.Impl(DataInputStream::readVector4f, DataOutputStream::writeStruct4f)

fun SynchedDelegates.vec2i(value: Vector2i, getter: DelegateGetter<Vector2i> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector2i> = DelegateSetter.passthrough()): ListenableDelegate<Vector2i> = RegularValue(value, Vector2iCodec, getter, setter)
fun SynchedDelegates.vec2d(value: Vector2d, getter: DelegateGetter<Vector2d> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector2d> = DelegateSetter.passthrough()): ListenableDelegate<Vector2d> = RegularValue(value, Vector2dCodec, getter, setter)
fun SynchedDelegates.vec2f(value: Vector2f, getter: DelegateGetter<Vector2f> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector2f> = DelegateSetter.passthrough()): ListenableDelegate<Vector2f> = RegularValue(value, Vector2fCodec, getter, setter)

fun SynchedDelegates.vec3i(value: Vector3i, getter: DelegateGetter<Vector3i> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector3i> = DelegateSetter.passthrough()): ListenableDelegate<Vector3i> = RegularValue(value, Vector3iCodec, getter, setter)
fun SynchedDelegates.vec3d(value: Vector3d, getter: DelegateGetter<Vector3d> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector3d> = DelegateSetter.passthrough()): ListenableDelegate<Vector3d> = RegularValue(value, Vector3dCodec, getter, setter)
fun SynchedDelegates.vec3f(value: Vector3f, getter: DelegateGetter<Vector3f> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector3f> = DelegateSetter.passthrough()): ListenableDelegate<Vector3f> = RegularValue(value, Vector3fCodec, getter, setter)

fun SynchedDelegates.vec4i(value: Vector4i, getter: DelegateGetter<Vector4i> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector4i> = DelegateSetter.passthrough()): ListenableDelegate<Vector4i> = RegularValue(value, Vector4iCodec, getter, setter)
fun SynchedDelegates.vec4d(value: Vector4d, getter: DelegateGetter<Vector4d> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector4d> = DelegateSetter.passthrough()): ListenableDelegate<Vector4d> = RegularValue(value, Vector4dCodec, getter, setter)
fun SynchedDelegates.vec4f(value: Vector4f, getter: DelegateGetter<Vector4f> = DelegateGetter.passthrough(), setter: DelegateSetter<Vector4f> = DelegateSetter.passthrough()): ListenableDelegate<Vector4f> = RegularValue(value, Vector4fCodec, getter, setter)
